@@ -38,20 +38,18 @@ import (
 	lbpb "google.golang.org/grpc/balancer/grpclb/grpc_lb_v1"
 	"google.golang.org/grpc/codes"
 	"google.golang.org/grpc/credentials"
-	_ "google.golang.org/grpc/grpclog/glogger"
-	"google.golang.org/grpc/internal/leakcheck"
+	"google.golang.org/grpc/internal/grpctest"
 	"google.golang.org/grpc/metadata"
 	"google.golang.org/grpc/peer"
 	"google.golang.org/grpc/resolver"
 	"google.golang.org/grpc/resolver/manual"
-	"google.golang.org/grpc/serviceconfig"
 	"google.golang.org/grpc/status"
 	testpb "google.golang.org/grpc/test/grpc_testing"
 )
 
 var (
-	lbServerName = "bar.com"
-	beServerName = "foo.com"
+	lbServerName = "lb.server.com"
+	beServerName = "backends.com"
 	lbToken      = "iamatoken"
 
 	// Resolver replaces localhost with fakeName in Next().
@@ -60,10 +58,17 @@ var (
 	fakeName = "fake.Name"
 )
 
+type s struct {
+	grpctest.Tester
+}
+
+func Test(t *testing.T) {
+	grpctest.RunSubTests(t, s{})
+}
+
 type serverNameCheckCreds struct {
-	mu       sync.Mutex
-	sn       string
-	expected string
+	mu sync.Mutex
+	sn string
 }
 
 func (c *serverNameCheckCreds) ServerHandshake(rawConn net.Conn) (net.Conn, credentials.AuthInfo, error) {
@@ -73,10 +78,10 @@ func (c *serverNameCheckCreds) ServerHandshake(rawConn net.Conn) (net.Conn, cred
 	}
 	return rawConn, nil, nil
 }
-func (c *serverNameCheckCreds) ClientHandshake(ctx context.Context, addr string, rawConn net.Conn) (net.Conn, credentials.AuthInfo, error) {
+func (c *serverNameCheckCreds) ClientHandshake(ctx context.Context, authority string, rawConn net.Conn) (net.Conn, credentials.AuthInfo, error) {
 	c.mu.Lock()
 	defer c.mu.Unlock()
-	b := make([]byte, len(c.expected))
+	b := make([]byte, len(authority))
 	errCh := make(chan error, 1)
 	go func() {
 		_, err := rawConn.Read(b)
@@ -85,34 +90,25 @@ func (c *serverNameCheckCreds) ClientHandshake(ctx context.Context, addr string,
 	select {
 	case err := <-errCh:
 		if err != nil {
-			fmt.Printf("Failed to read the server name from the server %v", err)
+			fmt.Printf("test-creds: failed to read expected authority name from the server: %v\n", err)
 			return nil, nil, err
 		}
 	case <-ctx.Done():
 		return nil, nil, ctx.Err()
 	}
-	if c.expected != string(b) {
-		fmt.Printf("Read the server name %s want %s", string(b), c.expected)
+	if authority != string(b) {
+		fmt.Printf("test-creds: got authority from ClientConn %q, expected by server %q\n", authority, string(b))
 		return nil, nil, errors.New("received unexpected server name")
 	}
 	return rawConn, nil, nil
 }
 func (c *serverNameCheckCreds) Info() credentials.ProtocolInfo {
-	c.mu.Lock()
-	defer c.mu.Unlock()
 	return credentials.ProtocolInfo{}
 }
 func (c *serverNameCheckCreds) Clone() credentials.TransportCredentials {
-	c.mu.Lock()
-	defer c.mu.Unlock()
-	return &serverNameCheckCreds{
-		expected: c.expected,
-	}
+	return &serverNameCheckCreds{}
 }
 func (c *serverNameCheckCreds) OverrideServerName(s string) error {
-	c.mu.Lock()
-	defer c.mu.Unlock()
-	c.expected = s
 	return nil
 }
 
@@ -177,18 +173,33 @@ func (s *rpcStats) equal(o *rpcStats) bool {
 	return mapsEqual(s.numCallsDropped, o.numCallsDropped)
 }
 
+func (s *rpcStats) String() string {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+	return fmt.Sprintf("Started: %v, Finished: %v, FinishedWithClientFailedToSend: %v, FinishedKnownReceived: %v, Dropped: %v",
+		atomic.LoadInt64(&s.numCallsStarted),
+		atomic.LoadInt64(&s.numCallsFinished),
+		atomic.LoadInt64(&s.numCallsFinishedWithClientFailedToSend),
+		atomic.LoadInt64(&s.numCallsFinishedKnownReceived),
+		s.numCallsDropped)
+}
+
 type remoteBalancer struct {
 	sls       chan *lbpb.ServerList
 	statsDura time.Duration
 	done      chan struct{}
 	stats     *rpcStats
+	statsChan chan *lbpb.ClientStats
+	fbChan    chan struct{}
 }
 
-func newRemoteBalancer(intervals []time.Duration) *remoteBalancer {
+func newRemoteBalancer(intervals []time.Duration, statsChan chan *lbpb.ClientStats) *remoteBalancer {
 	return &remoteBalancer{
-		sls:   make(chan *lbpb.ServerList, 1),
-		done:  make(chan struct{}),
-		stats: newRPCStats(),
+		sls:       make(chan *lbpb.ServerList, 1),
+		done:      make(chan struct{}),
+		stats:     newRPCStats(),
+		statsChan: statsChan,
+		fbChan:    make(chan struct{}),
 	}
 }
 
@@ -197,6 +208,10 @@ func (b *remoteBalancer) stop() {
 	close(b.done)
 }
 
+func (b *remoteBalancer) fallbackNow() {
+	b.fbChan <- struct{}{}
+}
+
 func (b *remoteBalancer) BalanceLoad(stream lbgrpc.LoadBalancer_BalanceLoadServer) error {
 	req, err := stream.Recv()
 	if err != nil {
@@ -229,6 +244,9 @@ func (b *remoteBalancer) BalanceLoad(stream lbgrpc.LoadBalancer_BalanceLoadServe
 				return
 			}
 			b.stats.merge(req.GetClientStats())
+			if b.statsChan != nil && req.GetClientStats() != nil {
+				b.statsChan <- req.GetClientStats()
+			}
 		}
 	}()
 	for {
@@ -239,6 +257,12 @@ func (b *remoteBalancer) BalanceLoad(stream lbgrpc.LoadBalancer_BalanceLoadServe
 					ServerList: v,
 				},
 			}
+		case <-b.fbChan:
+			resp = &lbpb.LoadBalanceResponse{
+				LoadBalanceResponseType: &lbpb.LoadBalanceResponse_FallbackResponse{
+					FallbackResponse: &lbpb.FallbackResponse{},
+				},
+			}
 		case <-stream.Context().Done():
 			return stream.Context().Err()
 		}
@@ -249,7 +273,7 @@ func (b *remoteBalancer) BalanceLoad(stream lbgrpc.LoadBalancer_BalanceLoadServe
 }
 
 type testServer struct {
-	testpb.TestServiceServer
+	testpb.UnimplementedTestServiceServer
 
 	addr     string
 	fallback bool
@@ -306,7 +330,7 @@ type testServers struct {
 	beListeners []net.Listener
 }
 
-func newLoadBalancer(numberOfBackends int) (tss *testServers, cleanup func(), err error) {
+func newLoadBalancer(numberOfBackends int, statsChan chan *lbpb.ClientStats) (tss *testServers, cleanup func(), err error) {
 	var (
 		beListeners []net.Listener
 		ls          *remoteBalancer
@@ -339,7 +363,7 @@ func newLoadBalancer(numberOfBackends int) (tss *testServers, cleanup func(), er
 		sn: lbServerName,
 	}
 	lb = grpc.NewServer(grpc.Creds(lbCreds))
-	ls = newRemoteBalancer(nil)
+	ls = newRemoteBalancer(nil, statsChan)
 	lbgrpc.RegisterLoadBalancerServer(lb, ls)
 	go func() {
 		lb.Serve(lbLis)
@@ -366,13 +390,11 @@ func newLoadBalancer(numberOfBackends int) (tss *testServers, cleanup func(), er
 	return
 }
 
-func TestGRPCLB(t *testing.T) {
-	defer leakcheck.Check(t)
-
+func (s) TestGRPCLB(t *testing.T) {
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
-	tss, cleanup, err := newLoadBalancer(1)
+	tss, cleanup, err := newLoadBalancer(1, nil)
 	if err != nil {
 		t.Fatalf("failed to create new load balancer: %v", err)
 	}
@@ -389,9 +411,7 @@ func TestGRPCLB(t *testing.T) {
 		Servers: bes,
 	}
 	tss.ls.sls <- sl
-	creds := serverNameCheckCreds{
-		expected: beServerName,
-	}
+	creds := serverNameCheckCreds{}
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
 	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
@@ -414,13 +434,11 @@ func TestGRPCLB(t *testing.T) {
 }
 
 // The remote balancer sends response with duplicates to grpclb client.
-func TestGRPCLBWeighted(t *testing.T) {
-	defer leakcheck.Check(t)
-
+func (s) TestGRPCLBWeighted(t *testing.T) {
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
-	tss, cleanup, err := newLoadBalancer(2)
+	tss, cleanup, err := newLoadBalancer(2, nil)
 	if err != nil {
 		t.Fatalf("failed to create new load balancer: %v", err)
 	}
@@ -440,9 +458,7 @@ func TestGRPCLBWeighted(t *testing.T) {
 		portsToIndex[tss.bePorts[i]] = i
 	}
 
-	creds := serverNameCheckCreds{
-		expected: beServerName,
-	}
+	creds := serverNameCheckCreds{}
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
 	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
@@ -484,13 +500,11 @@ func TestGRPCLBWeighted(t *testing.T) {
 	}
 }
 
-func TestDropRequest(t *testing.T) {
-	defer leakcheck.Check(t)
-
+func (s) TestDropRequest(t *testing.T) {
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
-	tss, cleanup, err := newLoadBalancer(2)
+	tss, cleanup, err := newLoadBalancer(2, nil)
 	if err != nil {
 		t.Fatalf("failed to create new load balancer: %v", err)
 	}
@@ -510,9 +524,7 @@ func TestDropRequest(t *testing.T) {
 			Drop: true,
 		}},
 	}
-	creds := serverNameCheckCreds{
-		expected: beServerName,
-	}
+	creds := serverNameCheckCreds{}
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
 	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
@@ -639,9 +651,7 @@ func TestDropRequest(t *testing.T) {
 }
 
 // When the balancer in use disconnects, grpclb should connect to the next address from resolved balancer address list.
-func TestBalancerDisconnects(t *testing.T) {
-	defer leakcheck.Check(t)
-
+func (s) TestBalancerDisconnects(t *testing.T) {
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
@@ -650,7 +660,7 @@ func TestBalancerDisconnects(t *testing.T) {
 		lbs   []*grpc.Server
 	)
 	for i := 0; i < 2; i++ {
-		tss, cleanup, err := newLoadBalancer(1)
+		tss, cleanup, err := newLoadBalancer(1, nil)
 		if err != nil {
 			t.Fatalf("failed to create new load balancer: %v", err)
 		}
@@ -672,9 +682,7 @@ func TestBalancerDisconnects(t *testing.T) {
 		lbs = append(lbs, tss.lb)
 	}
 
-	creds := serverNameCheckCreds{
-		expected: beServerName,
-	}
+	creds := serverNameCheckCreds{}
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
 	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
@@ -718,16 +726,14 @@ func TestBalancerDisconnects(t *testing.T) {
 	t.Fatalf("No RPC sent to second backend after 1 second")
 }
 
-func TestFallback(t *testing.T) {
+func (s) TestFallback(t *testing.T) {
 	balancer.Register(newLBBuilderWithFallbackTimeout(100 * time.Millisecond))
 	defer balancer.Register(newLBBuilder())
 
-	defer leakcheck.Check(t)
-
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
-	tss, cleanup, err := newLoadBalancer(1)
+	tss, cleanup, err := newLoadBalancer(1, nil)
 	if err != nil {
 		t.Fatalf("failed to create new load balancer: %v", err)
 	}
@@ -753,9 +759,7 @@ func TestFallback(t *testing.T) {
 		Servers: bes,
 	}
 	tss.ls.sls <- sl
-	creds := serverNameCheckCreds{
-		expected: beServerName,
-	}
+	creds := serverNameCheckCreds{}
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
 	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
@@ -771,9 +775,8 @@ func TestFallback(t *testing.T) {
 		Type:       resolver.GRPCLB,
 		ServerName: lbServerName,
 	}, {
-		Addr:       beLis.Addr().String(),
-		Type:       resolver.Backend,
-		ServerName: beServerName,
+		Addr: beLis.Addr().String(),
+		Type: resolver.Backend,
 	}}})
 
 	var p peer.Peer
@@ -789,9 +792,8 @@ func TestFallback(t *testing.T) {
 		Type:       resolver.GRPCLB,
 		ServerName: lbServerName,
 	}, {
-		Addr:       beLis.Addr().String(),
-		Type:       resolver.Backend,
-		ServerName: beServerName,
+		Addr: beLis.Addr().String(),
+		Type: resolver.Backend,
 	}}})
 
 	var backendUsed bool
@@ -812,12 +814,17 @@ func TestFallback(t *testing.T) {
 	// Close backend and remote balancer connections, should use fallback.
 	tss.beListeners[0].(*restartableListener).stopPreviousConns()
 	tss.lbListener.(*restartableListener).stopPreviousConns()
-	time.Sleep(time.Second)
 
 	var fallbackUsed bool
-	for i := 0; i < 1000; i++ {
+	for i := 0; i < 2000; i++ {
 		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
-			t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
+			// Because we are hard-closing the connection, above, it's possible
+			// for the first RPC attempt to be sent on the old connection,
+			// which will lead to an Unavailable error when it is closed.
+			// Ignore unavailable errors.
+			if status.Code(err) != codes.Unavailable {
+				t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
+			}
 		}
 		if p.Addr.String() == beLis.Addr().String() {
 			fallbackUsed = true
@@ -826,7 +833,7 @@ func TestFallback(t *testing.T) {
 		time.Sleep(time.Millisecond)
 	}
 	if !fallbackUsed {
-		t.Fatalf("No RPC sent to fallback after 1 second")
+		t.Fatalf("No RPC sent to fallback after 2 seconds")
 	}
 
 	// Restart backend and remote balancer, should not use backends.
@@ -834,10 +841,8 @@ func TestFallback(t *testing.T) {
 	tss.lbListener.(*restartableListener).restart()
 	tss.ls.sls <- sl
 
-	time.Sleep(time.Second)
-
 	var backendUsed2 bool
-	for i := 0; i < 1000; i++ {
+	for i := 0; i < 2000; i++ {
 		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
 			t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
 		}
@@ -848,23 +853,238 @@ func TestFallback(t *testing.T) {
 		time.Sleep(time.Millisecond)
 	}
 	if !backendUsed2 {
-		t.Fatalf("No RPC sent to backend behind remote balancer after 1 second")
+		t.Fatalf("No RPC sent to backend behind remote balancer after 2 seconds")
 	}
 }
 
-func TestGRPCLBPickFirst(t *testing.T) {
-	const pfc = `{"loadBalancingConfig":[{"grpclb":{"childPolicy":[{"pick_first":{}}]}}]}`
-	svcCfg, err := serviceconfig.Parse(pfc)
-	if err != nil {
-		t.Fatalf("Error parsing config %q: %v", pfc, err)
-	}
-
-	defer leakcheck.Check(t)
-
+func (s) TestExplicitFallback(t *testing.T) {
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
-	tss, cleanup, err := newLoadBalancer(3)
+	tss, cleanup, err := newLoadBalancer(1, nil)
+	if err != nil {
+		t.Fatalf("failed to create new load balancer: %v", err)
+	}
+	defer cleanup()
+
+	// Start a standalone backend.
+	beLis, err := net.Listen("tcp", "localhost:0")
+	if err != nil {
+		t.Fatalf("Failed to listen %v", err)
+	}
+	defer beLis.Close()
+	standaloneBEs := startBackends(beServerName, true, beLis)
+	defer stopBackends(standaloneBEs)
+
+	be := &lbpb.Server{
+		IpAddress:        tss.beIPs[0],
+		Port:             int32(tss.bePorts[0]),
+		LoadBalanceToken: lbToken,
+	}
+	var bes []*lbpb.Server
+	bes = append(bes, be)
+	sl := &lbpb.ServerList{
+		Servers: bes,
+	}
+	tss.ls.sls <- sl
+	creds := serverNameCheckCreds{}
+	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
+	defer cancel()
+	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
+		grpc.WithTransportCredentials(&creds), grpc.WithContextDialer(fakeNameDialer))
+	if err != nil {
+		t.Fatalf("Failed to dial to the backend %v", err)
+	}
+	defer cc.Close()
+	testC := testpb.NewTestServiceClient(cc)
+
+	r.UpdateState(resolver.State{Addresses: []resolver.Address{{
+		Addr:       tss.lbAddr,
+		Type:       resolver.GRPCLB,
+		ServerName: lbServerName,
+	}, {
+		Addr: beLis.Addr().String(),
+		Type: resolver.Backend,
+	}}})
+
+	var p peer.Peer
+	var backendUsed bool
+	for i := 0; i < 2000; i++ {
+		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
+			t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
+		}
+		if p.Addr.(*net.TCPAddr).Port == tss.bePorts[0] {
+			backendUsed = true
+			break
+		}
+		time.Sleep(time.Millisecond)
+	}
+	if !backendUsed {
+		t.Fatalf("No RPC sent to backend behind remote balancer after 2 seconds")
+	}
+
+	// Send fallback signal from remote balancer; should use fallback.
+	tss.ls.fallbackNow()
+
+	var fallbackUsed bool
+	for i := 0; i < 2000; i++ {
+		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
+			t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
+		}
+		if p.Addr.String() == beLis.Addr().String() {
+			fallbackUsed = true
+			break
+		}
+		time.Sleep(time.Millisecond)
+	}
+	if !fallbackUsed {
+		t.Fatalf("No RPC sent to fallback after 2 seconds")
+	}
+
+	// Send another server list; should use backends again.
+	tss.ls.sls <- sl
+
+	backendUsed = false
+	for i := 0; i < 2000; i++ {
+		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
+			t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
+		}
+		if p.Addr.(*net.TCPAddr).Port == tss.bePorts[0] {
+			backendUsed = true
+			break
+		}
+		time.Sleep(time.Millisecond)
+	}
+	if !backendUsed {
+		t.Fatalf("No RPC sent to backend behind remote balancer after 2 seconds")
+	}
+}
+
+func (s) TestFallBackWithNoServerAddress(t *testing.T) {
+	resolveNowCh := make(chan struct{}, 1)
+	r, cleanup := manual.GenerateAndRegisterManualResolver()
+	r.ResolveNowCallback = func(resolver.ResolveNowOptions) {
+		select {
+		case <-resolveNowCh:
+		default:
+		}
+		resolveNowCh <- struct{}{}
+	}
+	defer cleanup()
+
+	tss, cleanup, err := newLoadBalancer(1, nil)
+	if err != nil {
+		t.Fatalf("failed to create new load balancer: %v", err)
+	}
+	defer cleanup()
+
+	// Start a standalone backend.
+	beLis, err := net.Listen("tcp", "localhost:0")
+	if err != nil {
+		t.Fatalf("Failed to listen %v", err)
+	}
+	defer beLis.Close()
+	standaloneBEs := startBackends(beServerName, true, beLis)
+	defer stopBackends(standaloneBEs)
+
+	be := &lbpb.Server{
+		IpAddress:        tss.beIPs[0],
+		Port:             int32(tss.bePorts[0]),
+		LoadBalanceToken: lbToken,
+	}
+	var bes []*lbpb.Server
+	bes = append(bes, be)
+	sl := &lbpb.ServerList{
+		Servers: bes,
+	}
+	creds := serverNameCheckCreds{}
+	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
+	defer cancel()
+	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
+		grpc.WithTransportCredentials(&creds), grpc.WithContextDialer(fakeNameDialer))
+	if err != nil {
+		t.Fatalf("Failed to dial to the backend %v", err)
+	}
+	defer cc.Close()
+	testC := testpb.NewTestServiceClient(cc)
+
+	// Select grpclb with service config.
+	const pfc = `{"loadBalancingConfig":[{"grpclb":{"childPolicy":[{"round_robin":{}}]}}]}`
+	scpr := r.CC.ParseServiceConfig(pfc)
+	if scpr.Err != nil {
+		t.Fatalf("Error parsing config %q: %v", pfc, scpr.Err)
+	}
+
+	for i := 0; i < 2; i++ {
+		// Send an update with only backend address. grpclb should enter fallback
+		// and use the fallback backend.
+		r.UpdateState(resolver.State{
+			Addresses: []resolver.Address{{
+				Addr: beLis.Addr().String(),
+				Type: resolver.Backend,
+			}},
+			ServiceConfig: scpr,
+		})
+
+		select {
+		case <-resolveNowCh:
+			t.Errorf("unexpected resolveNow when grpclb gets no balancer address 1111, %d", i)
+		case <-time.After(time.Second):
+		}
+
+		var p peer.Peer
+		rpcCtx, rpcCancel := context.WithTimeout(context.Background(), time.Second)
+		defer rpcCancel()
+		if _, err := testC.EmptyCall(rpcCtx, &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
+			t.Fatalf("_.EmptyCall(_, _) = _, %v, want _, <nil>", err)
+		}
+		if p.Addr.String() != beLis.Addr().String() {
+			t.Fatalf("got peer: %v, want peer: %v", p.Addr, beLis.Addr())
+		}
+
+		select {
+		case <-resolveNowCh:
+			t.Errorf("unexpected resolveNow when grpclb gets no balancer address 2222, %d", i)
+		case <-time.After(time.Second):
+		}
+
+		tss.ls.sls <- sl
+		// Send an update with balancer address. The backends behind grpclb should
+		// be used.
+		r.UpdateState(resolver.State{
+			Addresses: []resolver.Address{{
+				Addr:       tss.lbAddr,
+				Type:       resolver.GRPCLB,
+				ServerName: lbServerName,
+			}, {
+				Addr: beLis.Addr().String(),
+				Type: resolver.Backend,
+			}},
+			ServiceConfig: scpr,
+		})
+
+		var backendUsed bool
+		for i := 0; i < 1000; i++ {
+			if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true), grpc.Peer(&p)); err != nil {
+				t.Fatalf("%v.EmptyCall(_, _) = _, %v, want _, <nil>", testC, err)
+			}
+			if p.Addr.(*net.TCPAddr).Port == tss.bePorts[0] {
+				backendUsed = true
+				break
+			}
+			time.Sleep(time.Millisecond)
+		}
+		if !backendUsed {
+			t.Fatalf("No RPC sent to backend behind remote balancer after 1 second")
+		}
+	}
+}
+
+func (s) TestGRPCLBPickFirst(t *testing.T) {
+	r, cleanup := manual.GenerateAndRegisterManualResolver()
+	defer cleanup()
+
+	tss, cleanup, err := newLoadBalancer(3, nil)
 	if err != nil {
 		t.Fatalf("failed to create new load balancer: %v", err)
 	}
@@ -888,9 +1108,7 @@ func TestGRPCLBPickFirst(t *testing.T) {
 		portsToIndex[tss.bePorts[i]] = i
 	}
 
-	creds := serverNameCheckCreds{
-		expected: beServerName,
-	}
+	creds := serverNameCheckCreds{}
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
 	cc, err := grpc.DialContext(ctx, r.Scheme()+":///"+beServerName,
@@ -908,6 +1126,11 @@ func TestGRPCLBPickFirst(t *testing.T) {
 	tss.ls.sls <- &lbpb.ServerList{Servers: beServers[0:3]}
 
 	// Start with sub policy pick_first.
+	const pfc = `{"loadBalancingConfig":[{"grpclb":{"childPolicy":[{"pick_first":{}}]}}]}`
+	scpr := r.CC.ParseServiceConfig(pfc)
+	if scpr.Err != nil {
+		t.Fatalf("Error parsing config %q: %v", pfc, scpr.Err)
+	}
 
 	r.UpdateState(resolver.State{
 		Addresses: []resolver.Address{{
@@ -915,7 +1138,7 @@ func TestGRPCLBPickFirst(t *testing.T) {
 			Type:       resolver.GRPCLB,
 			ServerName: lbServerName,
 		}},
-		ServiceConfig: svcCfg,
+		ServiceConfig: scpr,
 	})
 
 	result = ""
@@ -954,9 +1177,9 @@ func TestGRPCLBPickFirst(t *testing.T) {
 	}
 
 	// Switch sub policy to roundrobin.
-	grpclbServiceConfigEmpty, err := serviceconfig.Parse(`{}`)
-	if err != nil {
-		t.Fatalf("Error parsing config %q: %v", grpclbServiceConfigEmpty, err)
+	grpclbServiceConfigEmpty := r.CC.ParseServiceConfig(`{}`)
+	if grpclbServiceConfigEmpty.Err != nil {
+		t.Fatalf("Error parsing config %q: %v", `{}`, grpclbServiceConfigEmpty.Err)
 	}
 
 	r.UpdateState(resolver.State{
@@ -1012,13 +1235,11 @@ func checkStats(stats, expected *rpcStats) error {
 	return nil
 }
 
-func runAndGetStats(t *testing.T, drop bool, runRPCs func(*grpc.ClientConn)) *rpcStats {
-	defer leakcheck.Check(t)
-
+func runAndCheckStats(t *testing.T, drop bool, statsChan chan *lbpb.ClientStats, runRPCs func(*grpc.ClientConn), statsWant *rpcStats) error {
 	r, cleanup := manual.GenerateAndRegisterManualResolver()
 	defer cleanup()
 
-	tss, cleanup, err := newLoadBalancer(1)
+	tss, cleanup, err := newLoadBalancer(1, statsChan)
 	if err != nil {
 		t.Fatalf("failed to create new load balancer: %v", err)
 	}
@@ -1036,7 +1257,7 @@ func runAndGetStats(t *testing.T, drop bool, runRPCs func(*grpc.ClientConn)) *rp
 	}
 	tss.ls.sls <- &lbpb.ServerList{Servers: servers}
 	tss.ls.statsDura = 100 * time.Millisecond
-	creds := serverNameCheckCreds{expected: beServerName}
+	creds := serverNameCheckCreds{}
 
 	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
 	defer cancel()
@@ -1056,9 +1277,14 @@ func runAndGetStats(t *testing.T, drop bool, runRPCs func(*grpc.ClientConn)) *rp
 	}}})
 
 	runRPCs(cc)
-	time.Sleep(1 * time.Second)
-	stats := tss.ls.stats
-	return stats
+	end := time.Now().Add(time.Second)
+	for time.Now().Before(end) {
+		if err := checkStats(tss.ls.stats, statsWant); err == nil {
+			time.Sleep(200 * time.Millisecond) // sleep for two intervals to make sure no new stats are reported.
+			break
+		}
+	}
+	return checkStats(tss.ls.stats, statsWant)
 }
 
 const (
@@ -1066,9 +1292,8 @@ const (
 	failtosendURI = "failtosend"
 )
 
-func TestGRPCLBStatsUnarySuccess(t *testing.T) {
-	defer leakcheck.Check(t)
-	stats := runAndGetStats(t, false, func(cc *grpc.ClientConn) {
+func (s) TestGRPCLBStatsUnarySuccess(t *testing.T) {
+	if err := runAndCheckStats(t, false, nil, func(cc *grpc.ClientConn) {
 		testC := testpb.NewTestServiceClient(cc)
 		// The first non-failfast RPC succeeds, all connections are up.
 		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true)); err != nil {
@@ -1077,9 +1302,7 @@ func TestGRPCLBStatsUnarySuccess(t *testing.T) {
 		for i := 0; i < countRPC-1; i++ {
 			testC.EmptyCall(context.Background(), &testpb.Empty{})
 		}
-	})
-
-	if err := checkStats(stats, &rpcStats{
+	}, &rpcStats{
 		numCallsStarted:               int64(countRPC),
 		numCallsFinished:              int64(countRPC),
 		numCallsFinishedKnownReceived: int64(countRPC),
@@ -1088,9 +1311,8 @@ func TestGRPCLBStatsUnarySuccess(t *testing.T) {
 	}
 }
 
-func TestGRPCLBStatsUnaryDrop(t *testing.T) {
-	defer leakcheck.Check(t)
-	stats := runAndGetStats(t, true, func(cc *grpc.ClientConn) {
+func (s) TestGRPCLBStatsUnaryDrop(t *testing.T) {
+	if err := runAndCheckStats(t, true, nil, func(cc *grpc.ClientConn) {
 		testC := testpb.NewTestServiceClient(cc)
 		// The first non-failfast RPC succeeds, all connections are up.
 		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true)); err != nil {
@@ -1099,9 +1321,7 @@ func TestGRPCLBStatsUnaryDrop(t *testing.T) {
 		for i := 0; i < countRPC-1; i++ {
 			testC.EmptyCall(context.Background(), &testpb.Empty{})
 		}
-	})
-
-	if err := checkStats(stats, &rpcStats{
+	}, &rpcStats{
 		numCallsStarted:               int64(countRPC),
 		numCallsFinished:              int64(countRPC),
 		numCallsFinishedKnownReceived: int64(countRPC) / 2,
@@ -1111,9 +1331,8 @@ func TestGRPCLBStatsUnaryDrop(t *testing.T) {
 	}
 }
 
-func TestGRPCLBStatsUnaryFailedToSend(t *testing.T) {
-	defer leakcheck.Check(t)
-	stats := runAndGetStats(t, false, func(cc *grpc.ClientConn) {
+func (s) TestGRPCLBStatsUnaryFailedToSend(t *testing.T) {
+	if err := runAndCheckStats(t, false, nil, func(cc *grpc.ClientConn) {
 		testC := testpb.NewTestServiceClient(cc)
 		// The first non-failfast RPC succeeds, all connections are up.
 		if _, err := testC.EmptyCall(context.Background(), &testpb.Empty{}, grpc.WaitForReady(true)); err != nil {
@@ -1122,21 +1341,18 @@ func TestGRPCLBStatsUnaryFailedToSend(t *testing.T) {
 		for i := 0; i < countRPC-1; i++ {
 			cc.Invoke(context.Background(), failtosendURI, &testpb.Empty{}, nil)
 		}
-	})
-
-	if err := checkStats(stats, &rpcStats{
-		numCallsStarted:                        int64(countRPC),
-		numCallsFinished:                       int64(countRPC),
-		numCallsFinishedWithClientFailedToSend: int64(countRPC - 1),
+	}, &rpcStats{
+		numCallsStarted:                        int64(countRPC)*2 - 1,
+		numCallsFinished:                       int64(countRPC)*2 - 1,
+		numCallsFinishedWithClientFailedToSend: int64(countRPC-1) * 2,
 		numCallsFinishedKnownReceived:          1,
 	}); err != nil {
 		t.Fatal(err)
 	}
 }
 
-func TestGRPCLBStatsStreamingSuccess(t *testing.T) {
-	defer leakcheck.Check(t)
-	stats := runAndGetStats(t, false, func(cc *grpc.ClientConn) {
+func (s) TestGRPCLBStatsStreamingSuccess(t *testing.T) {
+	if err := runAndCheckStats(t, false, nil, func(cc *grpc.ClientConn) {
 		testC := testpb.NewTestServiceClient(cc)
 		// The first non-failfast RPC succeeds, all connections are up.
 		stream, err := testC.FullDuplexCall(context.Background(), grpc.WaitForReady(true))
@@ -1159,9 +1375,7 @@ func TestGRPCLBStatsStreamingSuccess(t *testing.T) {
 				}
 			}
 		}
-	})
-
-	if err := checkStats(stats, &rpcStats{
+	}, &rpcStats{
 		numCallsStarted:               int64(countRPC),
 		numCallsFinished:              int64(countRPC),
 		numCallsFinishedKnownReceived: int64(countRPC),
@@ -1170,9 +1384,8 @@ func TestGRPCLBStatsStreamingSuccess(t *testing.T) {
 	}
 }
 
-func TestGRPCLBStatsStreamingDrop(t *testing.T) {
-	defer leakcheck.Check(t)
-	stats := runAndGetStats(t, true, func(cc *grpc.ClientConn) {
+func (s) TestGRPCLBStatsStreamingDrop(t *testing.T) {
+	if err := runAndCheckStats(t, true, nil, func(cc *grpc.ClientConn) {
 		testC := testpb.NewTestServiceClient(cc)
 		// The first non-failfast RPC succeeds, all connections are up.
 		stream, err := testC.FullDuplexCall(context.Background(), grpc.WaitForReady(true))
@@ -1195,9 +1408,7 @@ func TestGRPCLBStatsStreamingDrop(t *testing.T) {
 				}
 			}
 		}
-	})
-
-	if err := checkStats(stats, &rpcStats{
+	}, &rpcStats{
 		numCallsStarted:               int64(countRPC),
 		numCallsFinished:              int64(countRPC),
 		numCallsFinishedKnownReceived: int64(countRPC) / 2,
@@ -1207,9 +1418,8 @@ func TestGRPCLBStatsStreamingDrop(t *testing.T) {
 	}
 }
 
-func TestGRPCLBStatsStreamingFailedToSend(t *testing.T) {
-	defer leakcheck.Check(t)
-	stats := runAndGetStats(t, false, func(cc *grpc.ClientConn) {
+func (s) TestGRPCLBStatsStreamingFailedToSend(t *testing.T) {
+	if err := runAndCheckStats(t, false, nil, func(cc *grpc.ClientConn) {
 		testC := testpb.NewTestServiceClient(cc)
 		// The first non-failfast RPC succeeds, all connections are up.
 		stream, err := testC.FullDuplexCall(context.Background(), grpc.WaitForReady(true))
@@ -1224,14 +1434,48 @@ func TestGRPCLBStatsStreamingFailedToSend(t *testing.T) {
 		for i := 0; i < countRPC-1; i++ {
 			cc.NewStream(context.Background(), &grpc.StreamDesc{}, failtosendURI)
 		}
-	})
-
-	if err := checkStats(stats, &rpcStats{
-		numCallsStarted:                        int64(countRPC),
-		numCallsFinished:                       int64(countRPC),
-		numCallsFinishedWithClientFailedToSend: int64(countRPC - 1),
+	}, &rpcStats{
+		numCallsStarted:                        int64(countRPC)*2 - 1,
+		numCallsFinished:                       int64(countRPC)*2 - 1,
+		numCallsFinishedWithClientFailedToSend: int64(countRPC-1) * 2,
 		numCallsFinishedKnownReceived:          1,
 	}); err != nil {
 		t.Fatal(err)
 	}
 }
+
+func (s) TestGRPCLBStatsQuashEmpty(t *testing.T) {
+	ch := make(chan *lbpb.ClientStats)
+	defer close(ch)
+	if err := runAndCheckStats(t, false, ch, func(cc *grpc.ClientConn) {
+		// Perform no RPCs; wait for load reports to start, which should be
+		// zero, then expect no other load report within 5x the update
+		// interval.
+		select {
+		case st := <-ch:
+			if !isZeroStats(st) {
+				t.Errorf("got stats %v; want all zero", st)
+			}
+		case <-time.After(5 * time.Second):
+			t.Errorf("did not get initial stats report after 5 seconds")
+			return
+		}
+
+		select {
+		case st := <-ch:
+			t.Errorf("got unexpected stats report: %v", st)
+		case <-time.After(500 * time.Millisecond):
+			// Success.
+		}
+		go func() {
+			for range ch { // Drain statsChan until it is closed.
+			}
+		}()
+	}, &rpcStats{
+		numCallsStarted:               0,
+		numCallsFinished:              0,
+		numCallsFinishedKnownReceived: 0,
+	}); err != nil {
+		t.Fatal(err)
+	}
+}
