@@ -27,8 +27,8 @@ package tools
 
 import (
 	_ "github.com/client9/misspell/cmd/misspell"
-	_ "github.com/golang/lint/golint"
 	_ "github.com/golang/protobuf/protoc-gen-go"
+	_ "golang.org/x/lint/golint"
 	_ "golang.org/x/tools/cmd/goimports"
 	_ "honnef.co/go/tools/cmd/staticcheck"
 )
